#include "mainwindow.h"
#include "ui_mainwindow.h"

#include <QFileDialog>
#include <QMessageBox>
#include <QRegExpValidator>

#include "PlotterWidget.hpp"

#include <QJsonDocument>
#include <QJsonArray>
#include <QJsonObject>

#include "utils.hpp"
#include "Objects/Well.hpp"
using namespace std;



MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    initPlot();

    QRegExp rx("-?\\d+\\.?\\d*");
    QValidator * validator = new QRegExpValidator(rx);
    ui->azimuthVs->setValidator(validator);

    mPlotter->setAzimuthVS(getAzimuthRad());

}

void MainWindow::on_importTrajectory_clicked()
{
    QString fileName =  QFileDialog::getOpenFileName(this, "Trajectory Import", "", tr("Json файлы (*.json)"));
    QFile file(fileName);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
        return;

    QString jsonFile(file.readAll());
    if (jsonFile.isNull())
        return;

    mWell = Utils::parseJson(jsonFile);
    if (mWell == nullptr)
        return;

    Utils::calculateTrajectory(*mWell);
    Utils::calculateVerticalSectionProjection(*mWell, getAzimuthRad());

    mPlotter->drawTrajectory(*mWell);

    mPlotter->scaleByTrajectory(*mWell);
}

void MainWindow::on_importLog_clicked()
{
    QString fileName =  QFileDialog::getOpenFileName(this, "Log Import", "", tr("Json файлы (*.json)"));
    QFile file(fileName);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
        return;

    QString json(file.readAll());
    if (json.isNull())
        return;


    QJsonDocument document = QJsonDocument::fromJson(json.toUtf8());
    if (document.isNull())
        return;

    vector<pair<double, double>> points;

    QJsonArray data_array(document.array());
    const int pointsCount(data_array.size());
    for(int i(0); i < pointsCount; ++i)
    {
        QJsonValue data(data_array[i].toObject()["data"]);
        QJsonValue md(data_array[i].toObject()["md"]);
        if(data.isNull() || md.isNull())
        {
            return;
        }
        points.push_back(make_pair(data.toDouble(), md.toDouble()));
    }

    log_data = std::make_unique<Well>();
    /// Добавление начальной точки траектории
    //log_data->addPoint(TrajectoryPoint(0.0, 0.0, 0.0));
    int ind = 0;

    vector<TrajectoryPoint> buff;

    for(int i = 0; i < points.size();i++){
        while(ind + 1 < mWell->getSize() && mWell->getPoint(ind + 1).mMeasureDepth <= points[i].second){
            ind += 1;
        }
        if(ind + 1 >= mWell->getSize()){
            break;
        }

        //printf("%lf\n", (points[i].second - mWell->getPoint(ind).mMeasureDepth) / (mWell->getPoint(ind + 1).mMeasureDepth - mWell->getPoint(ind).mMeasureDepth));
        log_data->addPoint(Utils::calculateTrajectoryPointPositionParticly(mWell->getPoint(ind), mWell->getPoint(ind + 1), (points[i].second - mWell->getPoint(ind).mMeasureDepth) / (mWell->getPoint(ind + 1).mMeasureDepth - mWell->getPoint(ind).mMeasureDepth)));
        buff.push_back(Utils::calculateTrajectoryPointPositionParticly(mWell->getPoint(ind), mWell->getPoint(ind + 1), (points[i].second - mWell->getPoint(ind).mMeasureDepth) / (mWell->getPoint(ind + 1).mMeasureDepth - mWell->getPoint(ind).mMeasureDepth)));
        log_data->getPoint(i).upByNormal(mWell->getPoint(ind), mWell->getPoint(ind + 1), points[i].first);

    }

    for(int i = buff.size() - 1; i >= 0;i--){
        log_data->addPoint(buff[i]);
    }
    log_data->addPoint(log_data->getPoint(0));
    Utils::calculateVerticalSectionProjection(*log_data, getAzimuthRad());

    mPlotter->drawCurve(*log_data);

}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::on_exit_clicked()
{
    this->close();
}

void MainWindow::on_azimuthVs_returnPressed()
{
    if(mWell != nullptr)
        Utils::calculateVerticalSectionProjection(*mWell, getAzimuthRad());
    if(log_data != nullptr)
        Utils::calculateVerticalSectionProjection(*log_data, getAzimuthRad());
    mPlotter->setAzimuthVS(getAzimuthRad());
    if(mWell != nullptr)
        mPlotter->drawTrajectory(*mWell);
    if(log_data != nullptr)
        mPlotter->drawCurve(*log_data);
}

void MainWindow::initPlot()
{
    mPlotter = new PlotterWidget(ui->centralWidget);
    QSizePolicy sizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding);
    sizePolicy.setHorizontalStretch(0);
    sizePolicy.setVerticalStretch(0);
    sizePolicy.setHeightForWidth(mPlotter->sizePolicy().hasHeightForWidth());
    mPlotter->setSizePolicy(sizePolicy);

    ui->horizontalLayout->insertWidget(0, mPlotter);
}

double MainWindow::getAzimuthRad() const
{
    return Utils::degreeToRadians(ui->azimuthVs->text().toDouble());
}
