#include "utils.hpp"

#include <stdexcept>

#include <QtMath>
#include <QVector3D>
#include <QJsonDocument>
#include <QJsonArray>
#include <QJsonObject>
using namespace std;
#include "Objects/Well.hpp"

namespace
{

constexpr double gPI(M_PI);

/**
 * Возвращает value, если оно лежит между start и end
 * Иначе возвращает ближайшее значение из двух
 */
template<typename T>
T clamp(T start, T end, T value)
{
    return std::max(start, std::min(end, value));
}

/// Вычисление длины вектора с координатами x, y
double calcClosureDistance(double x, double y)
{
    return std::sqrt(std::pow(x, 2) +
                     std::pow(y, 2));
}

/// Вычисление угла между вектором (x, y) и осью Y
double calcClosureDirectionRad(double x, double y)
{
    try
    {
        return std::atan2(x, y);
    }
    catch (std::domain_error &)
    {
        return 0.0;
    }
}

/**
 * Вычисление проекции точки (x, y) на вертикальную плоскость разреза
 * azimuthRad - угол между плоскостью и положительным направлением оси Y
 */

/// Нахождение нормализованного 3D вектора, заданного двумя 2D векторами в перпендикулярных
/// плоскостях
QVector3D makeUnitVectorBySurveyAngles(double azimuthRad, double inclinationRad)
{
    /// Для целей олимпиадной задачи потеря точности допустима
    QVector3D unitVector(static_cast<float>(std::sin(azimuthRad) * std::sin(inclinationRad)),
                         static_cast<float>(std::cos(azimuthRad) * std::sin(inclinationRad)),
                         static_cast<float>(std::cos(inclinationRad)));
    unitVector.normalize();
    return unitVector;
}

/// Вычисление значения кривизны участка траектории между двумя точками
double calcDogLeg(double prevInclinationRad, double inclinationRad,
                  double prevAzimutRad, double azimutRad)
{
    const auto u1 = makeUnitVectorBySurveyAngles(prevAzimutRad, prevInclinationRad);
    const auto u2 = makeUnitVectorBySurveyAngles(azimutRad, inclinationRad);

    const float cosValue = QVector3D::dotProduct(u1, u2);
    const double angleBetweenVectorsRad = std::acos(clamp(-1.0f, 1.0f, cosValue));

    return angleBetweenVectorsRad;
}

/// Расчёт мировых координат точки currentPoint по методу минимальной кривизны
void calculateTrajectoryPointPosition(const TrajectoryPoint & prevPoint,
                                      TrajectoryPoint & currentPoint)
{
    const double deltaMeasureDepth(currentPoint.mMeasureDepth - prevPoint.mMeasureDepth);
    const double dogLegRad(calcDogLeg(prevPoint.mInclination, currentPoint.mInclination,
                                      prevPoint.mAzimuth, currentPoint.mAzimuth));

    /**
     * Координаты касательного единичного вектора в первой точке
     */
    const double x1 = std::sin(prevPoint.mAzimuth) * std::sin(prevPoint.mInclination);
    const double y1 = std::cos(prevPoint.mAzimuth) * std::sin(prevPoint.mInclination);
    const double z1 = std::cos(prevPoint.mInclination);

    /**
     * Координаты касательного единичного вектора во второй точке
     */
    const double x2 = std::sin(currentPoint.mAzimuth) * std::sin(currentPoint.mInclination);
    const double y2 = std::cos(currentPoint.mAzimuth) * std::sin(currentPoint.mInclination);
    const double z2 = std::cos(currentPoint.mInclination);

    /// Вычисление координат для сонаправленных векторов
    const bool areVectorsCodirectional = qFuzzyCompare(dogLegRad, 0.0);
    if (areVectorsCodirectional)
    {
        const double deltaX = deltaMeasureDepth * x1;
        const double deltaY = deltaMeasureDepth * y1;
        const double deltaTrueVerticalDepth = deltaMeasureDepth * z1;

        currentPoint.mX = prevPoint.mX + deltaX;
        currentPoint.mY = prevPoint.mY + deltaY;
        currentPoint.mTrueVerticalDepth = prevPoint.mTrueVerticalDepth + deltaTrueVerticalDepth;

        return;
    }

    /// Вычисление координат для противоположно правленных векторов
    const bool areVectorsOpposite = qFuzzyCompare(dogLegRad, gPI);
    if (areVectorsOpposite)
    {
        /// В тестовом наборе данных такой случай не встречается, для упращения понимания кода
        /// ситуация не обрабатывается
    }

    /// Метод минимальной кривизны основан на аппроксимации участка траектории между двумя точками
    /// дугой окружности. Длина дуги соответствует расстоянию между двумя точками траектории (MD).
    /// Центральный угол дуги полагается равным значению dogleg, вычисленному для этих точек. Метод
    /// позволяет вычислить смещение второй точки относительно предыдущей.
    ///
    /// Для вычисления расстояния между двумя точками используются два касательных вектора.
    /// Обозначим первую точку как P1, а вторую как P2. Касательные вектора в этих точках пересекутся
    /// в точке T. Сумма векторов (P1,T) и (T, P2) равна вектору (P1, P2) - искомому смещению.
    /// Мы можем найти отношение длины векторов (P1,T) и (T, P2) к длине полудуги используя
    /// единичную окружность. Для неё длина дуги равна значению центрального угла dogLegRad
    /// в радианах. Зная это отношение и реальную длину дуги deltaMeasureDepth вычисляем реальную
    /// длину векторов (P1,T) и (T, P2). Имея длину векторов и единичные касательные вектора
    /// в точках P1 и P2 можно найти вектор (P1, P2)
    const double k = 0.5 * deltaMeasureDepth * std::tan(0.5 * dogLegRad) / (0.5 * dogLegRad);

    const double deltaX = k * (x1 + x2);
    const double deltaY = k * (y1 + y2);
    const double deltaTrueVerticalDepth = k * (z1 + z2);

    currentPoint.mX = prevPoint.mX + deltaX;
    currentPoint.mY = prevPoint.mY + deltaY;
    currentPoint.mTrueVerticalDepth = prevPoint.mTrueVerticalDepth + deltaTrueVerticalDepth;
}




} // anonymous namespace

namespace Utils
{

TrajectoryPoint calculateTrajectoryPointPositionParticly(const TrajectoryPoint & prevPoint,
                                      TrajectoryPoint & currentPoint, double val)
{
    const double deltaMeasureDepth((currentPoint.mMeasureDepth - prevPoint.mMeasureDepth) * val);
    const double dogLegRad(calcDogLeg(prevPoint.mInclination, currentPoint.mInclination,
                                      prevPoint.mAzimuth, currentPoint.mAzimuth));

    /**
     * Координаты касательного единичного вектора в первой точке
     */
    const double x1 = std::sin(prevPoint.mAzimuth) * std::sin(prevPoint.mInclination);
    const double y1 = std::cos(prevPoint.mAzimuth) * std::sin(prevPoint.mInclination);
    const double z1 = std::cos(prevPoint.mInclination);

    /**
     * Координаты касательного единичного вектора во второй точке
     */
    const double x2 = std::sin(currentPoint.mAzimuth ) * std::sin(currentPoint.mInclination * val);
    const double y2 = std::cos(currentPoint.mAzimuth ) * std::sin(currentPoint.mInclination * val);
    const double z2 = std::cos(currentPoint.mInclination * val);

    TrajectoryPoint result(prevPoint.mMeasureDepth + deltaMeasureDepth, currentPoint.mInclination * val, currentPoint.mAzimuth * val);

    const bool areVectorsCodirectional = qFuzzyCompare(dogLegRad, 0.0);
    if (areVectorsCodirectional)
    {
        const double deltaX = deltaMeasureDepth * x1;
        const double deltaY = deltaMeasureDepth * y1;
        const double deltaTrueVerticalDepth = deltaMeasureDepth * z1;

        result.mX = prevPoint.mX + deltaX;
        result.mY = prevPoint.mY + deltaY;
        result.mTrueVerticalDepth = prevPoint.mTrueVerticalDepth + deltaTrueVerticalDepth;

        return result;
    }

    const double k = 0.5 * deltaMeasureDepth * std::tan(0.5 * dogLegRad) / (0.5 * dogLegRad);

    const double deltaX = k * (x1 + x2);
    const double deltaY = k * (y1 + y2);
    const double deltaTrueVerticalDepth = k * (z1 + z2);

    result.mX = prevPoint.mX + deltaX;
    result.mY = prevPoint.mY + deltaY;
    result.mTrueVerticalDepth = prevPoint.mTrueVerticalDepth + deltaTrueVerticalDepth;
    return result;
}

/// преобразование градусов в радианы
double degreeToRadians(double degree)
{
    return gPI * degree / 180.0;
}

std::unique_ptr<Well> parseJson(const QString & json)
{
    std::unique_ptr<Well> well(std::make_unique<Well>());
    /// Добавление начальной точки траектории
    well->addPoint(TrajectoryPoint(0.0, 0.0, 0.0));

    QJsonDocument document = QJsonDocument::fromJson(json.toUtf8());
    if (document.isNull())
        return nullptr;

    QJsonArray data(document.array());
    const int pointsCount(data.size());
    for(int i(0); i < pointsCount; ++i)
    {
        QJsonValue azimuth(data[i].toObject()["azimuth"]);
        QJsonValue inclination(data[i].toObject()["inclination"]);
        QJsonValue depth(data[i].toObject()["depth"]);
        if(azimuth.isNull() || inclination.isNull() || depth.isNull())
        {
            return nullptr;
        }

        well->addPoint(TrajectoryPoint(depth.toDouble(), inclination.toDouble(), azimuth.toDouble()));
    }

    return well;
}

std::unique_ptr<Well> parseJsonLog(const QString & json)
{
    std::unique_ptr<Well> well(std::make_unique<Well>());
    /// Добавление начальной точки траектории
    well->addPoint(TrajectoryPoint(0.0, 0.0, 0.0));

    QJsonDocument document = QJsonDocument::fromJson(json.toUtf8());
    if (document.isNull())
        return nullptr;

    QJsonArray data(document.array());
    const int pointsCount(data.size());
    for(int i(0); i < pointsCount; ++i)
    {
        QJsonValue azimuth(data[i].toObject()["azimuth"]);
        QJsonValue inclination(data[i].toObject()["inclination"]);
        QJsonValue depth(data[i].toObject()["depth"]);
        if(azimuth.isNull() || inclination.isNull() || depth.isNull())
        {
            return nullptr;
        }

        well->addPoint(TrajectoryPoint(depth.toDouble(), inclination.toDouble(), azimuth.toDouble()));
    }

    return well;
}

void calculateTrajectory(Well & well)
{
    std::size_t pointsCount(well.getSize());
    /// Возможно рассчитать скважину только когда точек больше 1
    Q_ASSERT(pointsCount > 1);
    vector<TrajectoryPoint> res;
    /// Первая точка уже рассчитана, все координаты для неё нулевые из условия задачи
    /// Рассчитываем остальные точки по методу минимальной кривизны
    res.push_back(well.getPoint(0));
    for (std::size_t i(1); i < pointsCount; ++i)
    {
        const TrajectoryPoint & prevPoint(res[res.size() - 1]);
        TrajectoryPoint & currentPoint(well.getPoint(i));
        vector<TrajectoryPoint> bf;
        for(int g = 1; g <= 1;g++){
            //printf("%lf\n", (double)(g) / (double)(10.0));
            bf.push_back(Utils::calculateTrajectoryPointPositionParticly(prevPoint, currentPoint, (double)(g) / (double)(1.0)));
        }
        for(int g =0; g < bf.size();g++){
            res.push_back(bf[g]);
        } // некогда было нормально делать
        //calculateTrajectoryPointPosition(prevPoint, currentPoint);
    }
    well.mPoints = res;
}

double calculateVerticalSection(double x, double y, double azimuthRad)
{
    double length(calcClosureDistance(x, y));
    double closureDirectionRad(calcClosureDirectionRad(x, y));
    return length * std::cos(azimuthRad - closureDirectionRad);
}


void calculateVerticalSectionProjection(Well & well, double azimuthRad)
{
    std::size_t pointsCount(well.getSize());

    for (std::size_t i(0); i < pointsCount; ++i)
    {
        TrajectoryPoint & currentPoint(well.getPoint(i));
        currentPoint.mAzimuthProjection = calculateVerticalSection(currentPoint.mX, currentPoint.mY, azimuthRad);
    }
}

} // namespace Utils
